# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslotest import base
import psutil

from monasca_setup.detection.plugins import ceph


MON_PROCESSES = [
    {'name': 'ceph-mon.mon0',
     'type': 'ceph-mon',
     'search_string': [
         '/usr/bin/ceph-mon --cluster ceph --id mon0 -f',
         '/usr/bin/ceph-mon --cluster ceph -f --id mon0',
         '/usr/bin/ceph-mon --id mon0 --cluster ceph -f',
         '/usr/bin/ceph-mon --id mon0 -f --cluster ceph',
         '/usr/bin/ceph-mon -f --cluster ceph --id mon0',
         '/usr/bin/ceph-mon -f --id mon0 --cluster ceph'
     ]},
    {'name': 'ceph1-mon.mon0',
     'type': 'ceph-mon',
     'search_string': [
         '/usr/bin/ceph-mon --cluster ceph1 --id mon0 -f',
         '/usr/bin/ceph-mon --cluster ceph1 -f --id mon0',
         '/usr/bin/ceph-mon --id mon0 --cluster ceph1 -f',
         '/usr/bin/ceph-mon --id mon0 -f --cluster ceph1',
         '/usr/bin/ceph-mon -f --cluster ceph1 --id mon0',
         '/usr/bin/ceph-mon -f --id mon0 --cluster ceph1'
     ]},
]

RGW_PROCESSES = [
    {'name': 'ceph-radosgw.rgw0',
     'type': 'ceph-radosgw',
     'search_string': [
         '/usr/bin/radosgw --cluster ceph --name client.rgw.rgw0 -f',
         '/usr/bin/radosgw --cluster ceph -f --name client.rgw.rgw0',
         '/usr/bin/radosgw --name client.rgw.rgw0 --cluster ceph -f',
         '/usr/bin/radosgw --name client.rgw.rgw0 -f --cluster ceph',
         '/usr/bin/radosgw -f --cluster ceph --name client.rgw.rgw0',
         '/usr/bin/radosgw -f --name client.rgw.rgw0 --cluster ceph'
     ]},
    {'name': 'ceph1-radosgw.rgw0',
     'type': 'ceph-radosgw',
     'search_string': [
         '/usr/bin/radosgw --cluster ceph1 --name client.rgw.rgw0 -f',
         '/usr/bin/radosgw --cluster ceph1 -f --name client.rgw.rgw0',
         '/usr/bin/radosgw --name client.rgw.rgw0 --cluster ceph1 -f',
         '/usr/bin/radosgw --name client.rgw.rgw0 -f --cluster ceph1',
         '/usr/bin/radosgw -f --cluster ceph1 --name client.rgw.rgw0',
         '/usr/bin/radosgw -f --name client.rgw.rgw0 --cluster ceph1'
     ]},
]


def mocked_service_config(*args, **kwargs):
    if args[1] == 'mon':
        return MON_PROCESSES
    elif args[1] == 'radosgw':
        return RGW_PROCESSES
    return []


class FakeProcess(object):
    cmdLine = None

    def as_dict(self, attrs=None):
        all_attrs = {'name': 'ceph',
                     'exe': FakeProcess.exe(),
                     'cmdline': FakeProcess.cmdline()}
        if attrs:
            for key in attrs:
                if key not in all_attrs:
                    all_attrs.pop(key, None)
        return all_attrs

    @staticmethod
    def exe():
        line = FakeProcess.cmdLine
        if not line:
            return None
        return line[0]

    @staticmethod
    def cmdline():
        return FakeProcess.cmdLine


class TestCephDetection(base.BaseTestCase):
    CLUSTERS = [
        {
            'cluster_name': 'ceph',
            'config_file': '/etc/ceph/ceph.conf'
        },
        {
            'cluster_name': 'ceph1',
            'config_file': '/etc/ceph/ceph1.conf'
        },
    ]

    def setUp(self):
        super(TestCephDetection, self).setUp()
        with mock.patch.object(ceph.Ceph, '_detect') as mock_detect:
            self._ceph = ceph.Ceph('ceph')
            self.assertTrue(mock_detect.called)

    def test_should_not_configure_if_no_process(self):
        FakeProcess.cmdLine = []
        self._detect(proc=True)
        self.assertFalse(self._ceph.available)

    def test_should_be_available_if_everything_matches(self):
        ceph_cmd = '/usr/bin/ceph-mon -f --cluster ceph --id mon0 --setuser' \
               ' ceph --setgroup ceph'
        FakeProcess.cmdLine = [ceph_cmd]
        self._detect()
        self.assertTrue(self._ceph.available)

    def test_build_search_string(self):
        executable = '/usr/bin/ceph-mon'
        args = ['--cluster ceph', '--id mon0', '-f']

        expected_strings = [
            '/usr/bin/ceph-mon --cluster ceph --id mon0 -f',
            '/usr/bin/ceph-mon --cluster ceph -f --id mon0',
            '/usr/bin/ceph-mon --id mon0 --cluster ceph -f',
            '/usr/bin/ceph-mon --id mon0 -f --cluster ceph',
            '/usr/bin/ceph-mon -f --cluster ceph --id mon0',
            '/usr/bin/ceph-mon -f --id mon0 --cluster ceph'
        ]

        search_strings = self._ceph._build_search_string(executable, args)
        self.assertEqual(expected_strings, search_strings)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.listdir', return_value=['ceph-mon0', 'ceph1-mon0'])
    def test_service_config(self, list_dir, path_exists):
        processes = self._ceph._service_config(self.CLUSTERS, 'mon')
        self.assertEqual(MON_PROCESSES, processes)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.listdir', return_value=['ceph-rgw.rgw0', 'ceph1-rgw.rgw0'])
    def test_radosgw_service_config(self, list_dir, path_exists):
        processes = self._ceph._service_config(self.CLUSTERS, 'radosgw')
        self.assertEqual(RGW_PROCESSES, processes)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.listdir', return_value=[])
    def test_build_config_with_no_ceph_conf(self, list_dir, path_exists):
        config = self._ceph.build_config()
        self.assertEqual({}, dict(config))

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.listdir', return_value=['ceph.conf', 'ceph1.conf'])
    def test_build_config_with_no_admin_key(self, list_dir, path_exists):
        self._ceph._service_config = mock.Mock(
            side_effect=mocked_service_config)

        processes = MON_PROCESSES + RGW_PROCESSES
        process_instances = list()

        for p in processes:
            instance = {
                'exact_match': False,
                'search_string': p['search_string'],
                'detailed': True,
                'name': p['name'],
                'dimensions':  {'component': p['type'], 'service': 'ceph'}
            }
            process_instances.append(instance)

        expected_config = {
            'process': {
                'init_config': None,
                'instances': process_instances,
            },
            'ceph': {
                'init_config': None,
                'instances': [{'cluster_name': 'ceph',
                               'collect_mon_metrics': False,
                               'collect_osd_metrics': False,
                               'collect_pool_metrics': False,
                               'collect_stats_metrics': False,
                               'collect_usage_metrics': False},
                              {'cluster_name': 'ceph1',
                               'collect_mon_metrics': False,
                               'collect_osd_metrics': False,
                               'collect_pool_metrics': False,
                               'collect_stats_metrics': False,
                               'collect_usage_metrics': False}]
            }
        }
        config = self._ceph.build_config()
        self.assertEqual(expected_config, dict(config))

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.listdir', return_value=['ceph.conf', 'ceph1.conf', 'ceph1.client.admin.keyring'])
    def test_build_config(self, list_dir, path_exists):
        self._ceph._service_config = mock.Mock(
            side_effect=mocked_service_config)

        processes = MON_PROCESSES + RGW_PROCESSES
        process_instances = list()

        for p in processes:
            instance = {
                'exact_match': False,
                'search_string': p['search_string'],
                'detailed': True,
                'name': p['name'],
                'dimensions':  {'component': p['type'], 'service': 'ceph'}
            }
            process_instances.append(instance)

        expected_config = {
            'process': {
                'init_config': None,
                'instances': process_instances,
            },
            'ceph': {
                'init_config': None,
                'instances': [{'cluster_name': 'ceph',
                               'collect_mon_metrics': False,
                               'collect_osd_metrics': False,
                               'collect_pool_metrics': False,
                               'collect_stats_metrics': False,
                               'collect_usage_metrics': False},
                              {'cluster_name': 'ceph1'}]
            }
        }
        config = self._ceph.build_config()
        self.assertEqual(expected_config, dict(config))

    def _detect(self, proc=False):
        self._ceph.available = False
        processes = [FakeProcess()] if not proc else []
        process_iter = mock.patch.object(psutil, 'process_iter',
                                         return_value=processes)
        with process_iter as mock_process_iter:
            self._ceph._detect()
            self.assertTrue(mock_process_iter.called)



# (C) Copyright 2015-2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Aggregation classes used by the collector and statsd to batch messages sent to the forwarder.
"""
import json
import logging
from time import time

import monasca_common.validation.metrics as metric_validator

log = logging.getLogger(__name__)

# This is used to ensure that metrics with a timestamp older than
# RECENT_POINT_THRESHOLD_DEFAULT seconds (or the value passed in to
# the MetricsAggregator constructor) get discarded rather than being
# input into the incorrect bucket. Currently, the MetricsAggregator
# does not support submitting values for the past, and all values get
# submitted for the timestamp passed into the flush() function.
RECENT_POINT_THRESHOLD_DEFAULT = 3600
VALUE_META_VALUE_MAX_LENGTH = 2048


class MetricsAggregator(object):
    """A metric aggregator class."""

    def __init__(self, hostname, recent_point_threshold=None, tenant_id=None):
        self.total_count = 0
        self.count = 0
        self.hostname = hostname
        self.global_delegated_tenant = tenant_id

        recent_point_threshold = recent_point_threshold or RECENT_POINT_THRESHOLD_DEFAULT
        self.recent_point_threshold = int(recent_point_threshold)
        self.num_discarded_old_points = 0

        self.metrics = {}

    def flush(self):
        # Flush samples.  The individual metrics reset their internal samples
        # when required
        metrics = []
        for context, metric in self.metrics.items():
            try:
                metrics.extend(metric.flush())
            except Exception:
                log.exception('Error flushing {0} metrics.'.format(metric.metric['name']))

        # Log a warning regarding metrics with old timestamps being submitted
        if self.num_discarded_old_points > 0:
            log.warn('{0} points were discarded as a result of having an old timestamp'.format(
                     self.num_discarded_old_points))
            self.num_discarded_old_points = 0

        # Save some stats.
        log.debug("received {0} payloads since last flush".format(self.count))
        self.total_count += self.count
        self.count = 0
        return metrics

    def get_hostname_to_post(self, hostname):
        if 'SUPPRESS' == hostname:
            return None
        return hostname or self.hostname

    def packets_per_second(self, interval):
        if interval == 0:
            return 0
        return round(float(self.count) / interval, 2)

    def submit_metric(self, name, value, metric_class, dimensions=None,
                      delegated_tenant=None, hostname=None, device_name=None,
                      value_meta=None, timestamp=None, sample_rate=1):
        # validate dimensions, name, value and value meta
        if dimensions:
            metric_validator.validate_dimensions(dimensions)

        metric_validator.validate_name(name)

        metric_validator.validate_value(value)

        if value_meta:
            metric_validator.validate_value_meta(value_meta)

        hostname_to_post = self.get_hostname_to_post(hostname)

        tenant_to_post = delegated_tenant or self.global_delegated_tenant

        dimensions_copy = dimensions.copy()

        if 'hostname' not in dimensions_copy and hostname_to_post:
            dimensions_copy.update({'hostname': hostname_to_post})

        # TODO(joe): Shouldn't device_name be added to dimensions in the check
        #            plugin?  Why is it special cased through so many layers?
        if device_name:
            dimensions_copy.update({'device': device_name})

        # TODO(joe): Decide if hostname_to_post and device_name are necessary
        #            for the context tuple
        context = (name, tuple(dimensions_copy.items()), tenant_to_post,
                   hostname_to_post, device_name)

        if context not in self.metrics:
            self.metrics[context] = metric_class(name,
                                                 dimensions_copy,
                                                 tenant=tenant_to_post)
        cur_time = time()
        if timestamp is not None:
            if cur_time - int(timestamp) > self.recent_point_threshold:
                log.debug(
                    "Discarding {0} - ts = {1}, current ts = {2} ".format(name,
                                                                          timestamp,
                                                                          cur_time))
                self.num_discarded_old_points += 1
                return
        else:
            timestamp = cur_time
        self.metrics[context].value_meta = value_meta
        self.metrics[context].sample(value, sample_rate, timestamp)


def get_value_meta_overage(value_meta):
    if len(json.dumps(value_meta)) > VALUE_META_VALUE_MAX_LENGTH:
        return len(json.dumps(value_meta)) - VALUE_META_VALUE_MAX_LENGTH
    return 0



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import re

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class HAProxy(monasca_setup.detection.Plugin):
    """Detect HAProxy daemon and setup configuration to monitor.
    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        if monasca_setup.detection.find_process_cmdline('haproxy') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        log.info("\tEnabling HAProxy process watching")

        config.merge(
            monasca_setup.detection.watch_process(
                ['haproxy'],
                'haproxy',
                exact_match=False))
        if monasca_setup.detection.find_process_cmdline('keepalived') is not None:
            config.merge(
                monasca_setup.detection.watch_process(
                    ['keepalived'],
                    'haproxy',
                    exact_match=False))

        proxy_cfgfile = '/etc/haproxy/haproxy.cfg'
        if os.path.exists(proxy_cfgfile):
            # parse the config file looking for the stats section and pulling out url/user/pass
            with open(proxy_cfgfile, 'r') as pcfg:
                proxy_cfg = pcfg.read()

            url = None
            user = None
            password = None
            for line in proxy_cfg.splitlines():
                if line.startswith('listen'):
                    listen_match = re.search(r'^listen.*stats\S*\s(.*)', line)
                    if listen_match is None:
                        continue
                    listen_socket = listen_match.group(1).split(':', 1)
                    if listen_socket[0] == '':
                        host = 'localhost'
                    else:
                        host = listen_socket[0].strip()
                    port = listen_socket[1].strip()
                    url = 'http://{0}:{1}'.format(host, port)
                if url is not None and line.strip().startswith('stats auth'):
                    auth = re.search(r'stats auth\s(.*)', line).group(1).split(':')
                    user = auth[0].strip()
                    password = auth[1].strip()

            if url is None:
                log.warn(
                    'Unable to parse haproxy config for stats url, skipping HAProxy check plugin'
                    'configuration')
            else:
                log.info('Enabling the HAProxy check plugin')
                instance_config = {
                    'name': url,
                    'url': url,
                    'status_check': False,
                    'collect_service_stats_only': True,
                    'collect_status_metrics': False}
                if user is not None:
                    instance_config['username'] = user
                if password is not None:
                    instance_config['password'] = password
                config['haproxy'] = {'init_config': None, 'instances': [instance_config]}

        return config

    def dependencies_installed(self):
        return True



# Copyright 2013 Cloudbase Solutions Srl
# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Implementation of Inspector abstraction for Hyper-V"""


from oslo_utils import units

from monasca_agent.collector.virt.hyperv import utilsv2
from monasca_agent.collector.virt import inspector as virt_inspector


def get_instance_name(instance):
    """Shortcut to get instance name."""
    return getattr(instance, 'OS-EXT-SRV-ATTR:instance_name', None)


class HyperVInspector(virt_inspector.Inspector):

    def __init__(self):
        super(HyperVInspector, self).__init__()
        self._utils = utilsv2.UtilsV2()

    def inspect_cpus(self, instance):
        instance_name = get_instance_name(instance)
        (cpu_clock_used,
         cpu_count, uptime) = self._utils.get_cpu_metrics(instance_name)
        host_cpu_clock, host_cpu_count = self._utils.get_host_cpu_info()

        cpu_percent_used = (cpu_clock_used /
                            float(host_cpu_clock * cpu_count))
        # Nanoseconds
        cpu_time = (int(uptime * cpu_percent_used) * units.k)

        return virt_inspector.CPUStats(number=cpu_count, time=cpu_time)

    def inspect_memory_usage(self, instance, duration=None):
        instance_name = get_instance_name(instance)
        usage = self._utils.get_memory_metrics(instance_name)
        return virt_inspector.MemoryUsageStats(usage=usage)

    def inspect_vnics(self, instance):
        instance_name = get_instance_name(instance)
        for vnic_metrics in self._utils.get_vnic_metrics(instance_name):
            interface = virt_inspector.Interface(
                name=vnic_metrics["element_name"],
                mac=vnic_metrics["address"],
                fref=None,
                parameters=None)

            stats = virt_inspector.InterfaceStats(
                rx_bytes=vnic_metrics['rx_mb'] * units.Mi,
                rx_packets=0,
                tx_bytes=vnic_metrics['tx_mb'] * units.Mi,
                tx_packets=0)

            yield (interface, stats)

    def inspect_disks(self, instance):
        instance_name = get_instance_name(instance)
        for disk_metrics in self._utils.get_disk_metrics(instance_name):
            disk = virt_inspector.Disk(device=disk_metrics['instance_id'])
            stats = virt_inspector.DiskStats(
                read_requests=0,
                # Return bytes
                read_bytes=disk_metrics['read_mb'] * units.Mi,
                write_requests=0,
                write_bytes=disk_metrics['write_mb'] * units.Mi,
                errors=0)

            yield (disk, stats)

    def inspect_disk_latency(self, instance):
        instance_name = get_instance_name(instance)
        for disk_metrics in self._utils.get_disk_latency_metrics(
                instance_name):
            disk = virt_inspector.Disk(device=disk_metrics['instance_id'])
            stats = virt_inspector.DiskLatencyStats(
                disk_latency=disk_metrics['disk_latency'])

            yield (disk, stats)

    def inspect_disk_iops(self, instance):
        instance_name = get_instance_name(instance)
        for disk_metrics in self._utils.get_disk_iops_count(instance_name):
            disk = virt_inspector.Disk(device=disk_metrics['instance_id'])
            stats = virt_inspector.DiskIOPSStats(
                iops_count=disk_metrics['iops_count'])

            yield (disk, stats)



# (C) Copyright 2017-2018 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import math
import requests
import time
import yaml

from prometheus_client.parser import text_string_to_metric_families

import monasca_agent.collector.checks as checks
import monasca_agent.collector.checks.utils as utils

KUBERNETES_LABELS = ['app']


class Prometheus(checks.AgentCheck):
    """Scrapes metrics from Prometheus endpoints
    Can be configured three ways:
        1. Autodetect endpoints by pod annotations
        2. Autodetect endpoints by services
        3. Manually configure each prometheus endpoints to scrape

    We autodetect based on the annotations assigned to pods/services.

    We look for the following entries:
    'prometheus.io/scrape': Only scrape pods that have a value of 'true'
    'prometheus.io/path': If the metrics path is not '/metrics' override this.
    'prometheus.io/port': Scrape the pod on the indicated port instead of the default of '9102'.

    Additional settings for prometheus endpoints
    'monasca.io/usek8slabels': Attach kubernetes labels of the pod that is being scraped.
        Default to 'true'
    'monasca.io/whitelist': Yaml list of metric names to whitelist against on detected endpoint
    'monasca.io/report_pod_label_owner': If the metrics that are scraped contain pod as a label
        key we will attempt to get the pod owner and attach that to the metric as another
        dimension. Very useful for other scraping from other solutions
        that monitor k8s (Ex. kube-state-metrics). Default to 'false'
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        super(Prometheus, self).__init__(name, init_config, agent_config, instances)
        self.connection_timeout = init_config.get("timeout", 3)
        self.auto_detect_endpoints = init_config.get("auto_detect_endpoints", False)
        if self.auto_detect_endpoints:
            self.kubernetes_connector = None
            self.detect_method = init_config.get("detect_method", "pod").lower()
            self.kubelet_url = None
            if instances is not None and len(instances) > 1:
                raise Exception(
                    'Prometheus Client only supports one configured instance if auto'
                    'detection is set')
            if self.detect_method not in ['pod', 'service']:
                raise Exception('Invalid detect method {}. Must be either pod or service')
            self.k8s_pod_cache = None
            self.cache_start_time = None

    def check(self, instance):
        dimensions = self._set_dimensions(None, instance)
        del dimensions['hostname']
        if not self.auto_detect_endpoints:
            metric_endpoint = instance.get("metric_endpoint", None)
            if not metric_endpoint:
                self.log.error("metric_endpoint must be defined for each instance")
                return
            endpoint_dimensions = instance.get("default_dimensions", {})
            endpoint_dimensions.update(dimensions)
            self.report_endpoint_metrics(metric_endpoint, endpoint_dimensions)
        else:
            self.kubernetes_labels = instance.get('kubernetes_labels', KUBERNETES_LABELS)
            if not self.kubernetes_connector:
                self.kubernetes_connector = utils.KubernetesConnector(self.connection_timeout)
            # Check if we need to clear pod cache so it does not build up over time
            if self.k8s_pod_cache is not None:
                if not self.cache_start_time:
                    self.cache_start_time = time.time()
                else:
                    current_time = time.time()
                    if (current_time - self.cache_start_time) > 86400:
                        self.cache_start_time = current_time
                        self.k8s_pod_cache = {}
                        self.initialize_pod_cache()
            if self.detect_method == "pod":
                if not self.kubelet_url:
                    try:
                        host = self.kubernetes_connector.get_agent_pod_host()
                        self.kubelet_url = "http://{}:10255/pods".format(host)
                    except Exception as e:
                        self.log.error("Could not obtain current host from Kubernetes API {}. "
                                       "Skipping check".format(e))
                        return
                prometheus_endpoints = self._get_metric_endpoints_by_pod(dimensions)
            # Detect by service
            else:
                prometheus_endpoints = self._get_metric_endpoints_by_service(dimensions)
            for prometheus_endpoint in prometheus_endpoints:
                self.report_endpoint_metrics(
                    prometheus_endpoint.scrape_endpoint,
                    prometheus_endpoint.dimensions,
                    prometheus_endpoint.whitelist,
                    prometheus_endpoint.metric_types,
                    prometheus_endpoint.report_pod_label_owner)

    def _get_metric_endpoints_by_pod(self, dimensions):
        prometheus_endpoints = []
        # Grab running pods from local Kubelet
        try:
            pods = requests.get(self.kubelet_url, timeout=self.connection_timeout).json()
        except Exception as e:
            exception_message = "Could not get pods from local kubelet with error - {}".format(e)
            self.log.exception(exception_message)
            raise Exception(exception_message)

        # Iterate through each pod and check if it contains a scrape endpoint
        for pod in pods['items']:
            try:
                pod_metadata = pod['metadata']
                pod_spec = pod['spec']
                pod_status = pod['status']
                if "annotations" not in pod_metadata or not (
                        'containers' in pod_spec and 'podIP' in pod_status):
                    # No annotations, containers, or endpoints skipping pod
                    continue

                # Check pod annotations if we should scrape pod
                pod_annotations = pod_metadata['annotations']
                prometheus_scrape = pod_annotations.get("prometheus.io/scrape", "false").lower()
                if prometheus_scrape != "true":
                    continue
                pod_ports = []
                pod_containers = pod_spec['containers']
                for container in pod_containers:
                    if "ports" in container:
                        pod_ports += container['ports']
                pod_name = pod_metadata['name']
                endpoints = self._get_prometheus_endpoint(pod_annotations, pod_ports, pod_name)
                if not endpoints:
                    continue

                # Add pod endpoint to scrape endpoints
                pod_ip = pod_status['podIP']
                # Loop through list of ports and build list of endpoints

                pod_dimensions = dimensions.copy()
                try:
                    use_k8s_labels, whitelist, metric_types, report_pod_label_owner = \
                        self._get_monasca_settings(pod_name, pod_annotations)
                except Exception as e:
                    error_message = "Error parsing monasca annotations on endpoints {} " \
                                    "with error - {}. " \
                                    "Skipping scraping metrics".format(endpoints, e)
                    self.log.error(error_message)
                    continue
                # set global_pod_cache
                if report_pod_label_owner and self.k8s_pod_cache is None:
                    self.k8s_pod_cache = {}
                    self.initialize_pod_cache()
                if use_k8s_labels:
                    pod_dimensions.update(utils.get_pod_dimensions(
                        self.kubernetes_connector, pod['metadata'],
                        self.kubernetes_labels))
                for endpoint in endpoints:
                    scrape_endpoint = "http://{}:{}".format(pod_ip, endpoint)
                    prometheus_endpoint = PrometheusEndpoint(
                        scrape_endpoint,
                        pod_dimensions,
                        whitelist,
                        metric_types,
                        report_pod_label_owner)
                    prometheus_endpoints.append(prometheus_endpoint)
                    self.log.info("Detected pod endpoint - {} with metadata "
                                  "of {}".format(scrape_endpoint,
                                                 pod_dimensions))
            except Exception as e:
                self.log.warn("Error parsing {} to detect for scraping - {}".format(pod, e))
                continue

        return prometheus_endpoints

    def _get_metric_endpoints_by_service(self, dimensions):
        prometheus_endpoints = []
        # Grab services from Kubernetes API
        try:
            services = self.kubernetes_connector.get_request("/api/v1/services")
        except Exception as e:
            exception_message = \
                "Could not get services from Kubernetes API with error - {}".format(e)
            self.log.exception(exception_message)
            raise Exception(exception_message)

        # Iterate through each service and check if it is a scape endpoint
        for service in services['items']:
            service_metadata = service['metadata']
            service_spec = service['spec']
            if "annotations" not in service_metadata or "ports" not in service_spec:
                # No annotations or pods skipping service
                continue

            # Check service annotations if we should scrape service
            service_annotations = service_metadata['annotations']
            prometheus_scrape = service_annotations.get("prometheus.io/scrape", "false").lower()
            if prometheus_scrape != "true":
                continue
            service_name = service_metadata['name']
            service_ports = service_spec['ports']
            endpoints = self._get_prometheus_endpoint(service_annotations,
                                                      service_ports,
                                                      service_name)
            if not endpoints:
                continue

            # Add service endpoint to scrape endpoints
            cluster_ip = service_spec['clusterIP']
            service_dimensions = dimensions.copy()
            try:
                use_k8s_labels, whitelist, metric_types, report_pod_label_owner = \
                    self._get_monasca_settings(service_name, service_annotations)
            except Exception as e:
                error_message = "Error parsing monasca annotations on endpoints {} " \
                                "with error - {}. " \
                                "Skipping scraping metrics".format(endpoints, e)
                self.log.error(error_message)
                continue
                # set global_pod_cache
            if report_pod_label_owner and self.k8s_pod_cache is None:
                self.k8s_pod_cache = {}
                self.initialize_pod_cache()
            if use_k8s_labels:
                service_dimensions.update(
                    self._get_service_dimensions(service_metadata))
            for endpoint in endpoints:
                scrape_endpoint = "http://{}:{}".format(cluster_ip, endpoint)
                prometheus_endpoint = PrometheusEndpoint(
                    scrape_endpoint,
                    service_dimensions,
                    whitelist,
                    metric_types,
                    report_pod_label_owner)
                prometheus_endpoints.append(prometheus_endpoint)
                self.log.info("Detected service endpoint - {} with metadata "
                              "of {}".format(scrape_endpoint,
                                             service_dimensions))
        return prometheus_endpoints

    def _get_monasca_settings(self, resource_name, annotations):
        use_k8s_labels = annotations.get("monasca.io/usek8slabels", "true").lower() == "true"
        whitelist = None
        if "monasca.io/whitelist" in annotations:
            whitelist = yaml.safe_load(annotations["monasca.io/whitelist"])
        metric_types = None
        if "monasca.io/metric_types" in annotations:
            metric_types = yaml.safe_load(annotations["monasca.io/metric_types"])
            for typ in metric_types:
                if metric_types[typ] not in ['rate', 'counter']:
                    self.log.warn(
                        "Ignoring unknown metric type '{}' configured for '{}'"
                        " on endpoint '{}'".format(
                            typ, metric_types[typ], resource_name))
                    del metric_types[typ]
        report_pod_label_owner_annotation = annotations.get(
            "monasca.io/report_pod_label_owner", "false").lower()
        report_pod_label_owner = True if report_pod_label_owner_annotation == "true" else False
        return use_k8s_labels, whitelist, metric_types, report_pod_label_owner

    def _get_service_dimensions(self, service_metadata):
        service_dimensions = {'service_name': service_metadata['name'],
                              'namespace': service_metadata['namespace']}
        if "labels" in service_metadata:
            service_labels = service_metadata['labels']
            for label in self.kubernetes_labels:
                if label in service_labels:
                    service_dimensions[label] = service_labels[label]
        return service_dimensions

    def _get_prometheus_endpoint(self, annotations, ports, name):
        """Analyzes annotations and ports to generate a scrape target"""
        pod_index = "containerPort" if self.detect_method == "pod" else "port"
        configured_ports = []
        if "prometheus.io/port" in annotations:
            configured_ports = annotations.get("prometheus.io/port").split(',')
            configured_ports = [int(i) for i in configured_ports]

        if self.detect_method == "pod" and not configured_ports:
            configured_ports = [9102]
        prometheus_endpoint = annotations.get("prometheus.io/path", "/metrics")
        prometheus_endpoint = prometheus_endpoint.lstrip('/')

        endpoints = []
        for port in ports:
            for configured_port in configured_ports:
                if port[pod_index] == configured_port:
                    # Build up list of ports and prometheus endpoints to return
                    endpoints.append("{}/{}".format(configured_port,
                                                    prometheus_endpoint))

        if len(ports) == 1 and not endpoints:
            self.log.info("Could not find matching port using only port "
                          "configured")
            endpoints.append("{}/{}".format(ports[0][pod_index], prometheus_endpoint))

        if not endpoints:
            self.log.error("Can not derive which port to use. Due to either "
                           "no port being exposed or more than one port "
                           "configured and none of them selected via "
                           "configurations. "
                           "{} {} skipped for scraping".format(self.detect_method, name))
        self.log.debug("Found prometheus endpoints '{}'".format(endpoints))
        return endpoints

    def _send_metrics(self, metric_families, dimensions, endpoint_whitelist, endpoint_metric_types,
                      report_pod_label_owner):
        for metric_family in metric_families:
            for metric in metric_family.samples:
                metric_dimensions = dimensions.copy()
                metric_name = metric[0]
                metric_labels = metric[1]
                metric_value = float(metric[2])
                if math.isnan(metric_value):
                    self.log.debug(
                        'filtering out NaN value provided for metric %s{%s}',
                        metric_name,
                        metric_labels)
                    continue
                if endpoint_whitelist is not None and metric_name not in endpoint_whitelist:
                    continue
                # remove empty string dimensions from prometheus labels
                for dim_key, dim_value in metric_labels.items():
                    if len(dim_value) > 0:
                        metric_dimensions[dim_key] = dim_value

                metric_func = self.gauge
                if endpoint_metric_types and metric_name in endpoint_metric_types:
                    typ = endpoint_metric_types[metric_name]
                    if typ == "rate":
                        metric_func = self.rate
                        metric_name += "_rate"
                    elif typ == "counter":
                        metric_func = self.increment
                        metric_name += "_counter"
                if report_pod_label_owner:
                    if "pod" in metric_dimensions and "namespace" in metric_dimensions:
                        pod_name = metric_dimensions["pod"]
                        if pod_name in self.k8s_pod_cache:
                            pod_owner, pod_owner_name = self.k8s_pod_cache[pod_name]
                            metric_dimensions[pod_owner] = pod_owner_name
                            metric_dimensions["owner_type"] = pod_owner
                        else:
                            pod_owner_pair = self.get_pod_owner(
                                pod_name, metric_dimensions['namespace'])
                            if pod_owner_pair:
                                pod_owner = pod_owner_pair[0]
                                pod_owner_name = pod_owner_pair[1]
                                metric_dimensions[pod_owner] = pod_owner_name
                                metric_dimensions["owner_type"] = pod_owner
                                self.k8s_pod_cache[pod_name] = pod_owner, pod_owner_name
                metric_func(
                    metric_name,
                    metric_value,
                    dimensions=metric_dimensions,
                    hostname="SUPPRESS")

    def report_endpoint_metrics(self, metric_endpoint, endpoint_dimensions, endpoint_whitelist=None,
                                endpoint_metric_types=None, report_pod_label_owner=False):
        # Hit metric endpoint
        try:
            result = requests.get(metric_endpoint, timeout=self.connection_timeout)
        except Exception as e:
            self.log.error("Could not get metrics from {} with error {}".format(metric_endpoint, e))
        else:
            result_content_type = result.headers['Content-Type']
            if "text/plain" in result_content_type:
                try:
                    metric_families = text_string_to_metric_families(result.text)
                    self._send_metrics(
                        metric_families,
                        endpoint_dimensions,
                        endpoint_whitelist,
                        endpoint_metric_types,
                        report_pod_label_owner)
                except Exception as e:
                    self.log.error(
                        "Error parsing data from {} with error {}".format(
                            metric_endpoint, e))
            else:
                self.log.error("Unsupported content type - {}".format(result_content_type))

    def get_pod_owner(self, pod_name, namespace):
        try:
            pod = self.kubernetes_connector.get_request(
                "/api/v1/namespaces/{}/pods/{}".format(namespace, pod_name))
            pod_metadata = pod['metadata']
            pod_owner, pod_owner_name = utils.get_pod_owner(self.kubernetes_connector, pod_metadata)
            if not pod_owner:
                self.log.info("Could not get pod owner for pod {}".format(pod_name))
                return None
            return pod_owner, pod_owner_name
        except Exception as e:
            self.log.info(
                "Could not get pod {} from Kubernetes API with error - {}".format(pod_name, e))
            return None

    def initialize_pod_cache(self):
        self.k8s_pod_cache = {}
        try:
            pods = self.kubernetes_connector.get_request("/api/v1/pods")
        except Exception as e:
            exception_message = "Could not get pods from Kubernetes API with error - {}".format(e)
            self.log.exception(exception_message)
            raise Exception(exception_message)
        for pod in pods['items']:
            pod_metadata = pod['metadata']
            pod_name = pod_metadata['name']
            try:
                pod_owner, pod_owner_name = utils.get_pod_owner(
                    self.kubernetes_connector, pod_metadata)
            except Exception as e:
                self.log.info(
                    "Error attempting to get pod {} owner with error {}".format(
                        pod_name, e))
                continue
            if not pod_owner:
                self.log.info("Could not get pod owner for pod {}".format(pod_name))
                continue
            self.k8s_pod_cache[pod_name] = (pod_owner, pod_owner_name)


# Class to hold prometheus endpoint metadata
class PrometheusEndpoint(object):
    def __init__(
            self,
            scrape_endpoint,
            dimensions,
            whitelist,
            metric_types,
            report_pod_label_owner):
        self.scrape_endpoint = scrape_endpoint
        self.dimensions = dimensions
        self.whitelist = whitelist
        self.metric_types = metric_types
        self.report_pod_label_owner = report_pod_label_owner



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
import socket
import ssl

from urllib.parse import urlparse

from monasca_agent.collector.checks import AgentCheck


class CertificateCheck(AgentCheck):
    """Uses ssl to get the SSL certificate and output a metric
       which is the number of days until it expires
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        super(CertificateCheck, self).__init__(name, init_config,
                                               agent_config, instances)
        self._ca_certs = init_config.get('ca_certs')
        self._ciphers = init_config.get('ciphers')
        self._timeout = init_config.get('timeout')
        self.log.debug('ca_certs file is %s' % self._ca_certs)
        self.log.debug('cipers are %s' % self._ciphers)
        self.log.debug('timeout is %f' % self._timeout)

    def check(self, instance):
        url = instance.get('url', None)
        dimensions = self._set_dimensions(None, instance)
        dimensions['url'] = url
        self.log.info('url = %s' % url)
        try:
            expire_date = self.get_expire(url, self._ca_certs, self._ciphers,
                                          self._timeout)
            expire_in = expire_date - datetime.now()
            self.gauge('https.cert_expire_days', expire_in.days,
                       dimensions=dimensions)
            self.log.debug('%d days till expiration for %s' % (expire_in.days,
                                                               url))

        except Exception as e:
            self.log.warning('Exception trying to GET certificate for %s: %s' %
                             (url, e))
            self.log.exception('Failed to get certificate for %s' % url)

    def get_expire(self, url, ca_certs, ciphers, timeout):
        url_info = urlparse(url)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        sock.connect((url_info.hostname, url_info.port))
        ssl_sock = ssl.wrap_socket(sock, cert_reqs=ssl.CERT_REQUIRED,
                                   ca_certs=ca_certs,
                                   ciphers=(ciphers))

        cert = ssl_sock.getpeercert()

        not_after = cert['notAfter']
        try:
            expire_date = datetime.strptime(not_after,
                                            '%b %d %H:%M:%S %Y %Z')
            return expire_date
        except ValueError:
            raise Exception('Invalid Certificate date format')



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import unittest

from monasca_setup.detection.plugins.cert_check import CertificateCheck

LOG = logging.getLogger('monasca_setup.detection.plugins.cert_check')


class TestCertCheck(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        with unittest.mock.patch.object(CertificateCheck, '_detect') as mock_detect:
            self.cert_obj = CertificateCheck('temp_dir')
            self.assertTrue(mock_detect.called)
            self.cert_obj.args = {'urls': 'http://fake-cert.com'}

    def test_detect(self):
        self.cert_obj.available = False
        with unittest.mock.patch.object(self.cert_obj, '_check_required_args',
                          return_value=True) as mock_check_required_args:
            self.cert_obj._detect()
            self.assertTrue(self.cert_obj.available)
            self.assertTrue(mock_check_required_args.called)

    def _build_config(self):
        with unittest.mock.patch.object(self.cert_obj, '_build_instance',
                          return_value={}) as mock_build_instance:
            result = self.cert_obj.build_config()
            self.assertTrue(mock_build_instance.called)
            self.assertEqual(result['cert_check']['instances'][0]['url'],
                             'http://fake-cert.com')
            self.assertEqual(result['cert_check']['instances'][0]['name'],
                             'http://fake-cert.com')
            return result

    def test_build_config_without_args(self):
        result = self._build_config()
        self.assertEqual(result['cert_check']['init_config']['ca_certs'],
                         '/etc/ssl/certs/ca-certificates.crt')
        self.assertEqual(result['cert_check']['init_config']['ciphers'],
                         'HIGH:-aNULL:-eNULL:-PSK:RC4-SHA:RC4-MD5')
        self.assertEqual(result['cert_check']['init_config']['collect_period'],
                         3600)
        self.assertEqual(result['cert_check']['init_config']['timeout'],
                         1.0)

    def test_build_config_with_args(self):
        self.cert_obj.args.update({'ca_certs': '/tmp/ssl/certs/ca-certificates.crt',
                                   'ciphers': 'fake-cipher',
                                   'timeout': 3.0,
                                   'collect_period': 1200})
        result = self._build_config()
        self.assertEqual(result['cert_check']['init_config']['ca_certs'],
                         '/tmp/ssl/certs/ca-certificates.crt')
        self.assertEqual(result['cert_check']['init_config']['ciphers'],
                         'fake-cipher')
        self.assertEqual(result['cert_check']['init_config']['collect_period'],
                         1200)
        self.assertEqual(result['cert_check']['init_config']['timeout'],
                         3.0)

    def test_timeout_error_log(self):
        self.cert_obj.args.update({'ca_certs': '/tmp/ssl/certs/ca-certificates.crt',
                                   'ciphers': 'fake-cipher',
                                   'timeout': 0.0,
                                   'collect_period': 1200})
        with unittest.mock.patch.object(LOG, 'error') as mock_log:
            result = self._build_config()
            self.assertEqual(result['cert_check']['init_config']['ca_certs'],
                             '/tmp/ssl/certs/ca-certificates.crt')
            self.assertEqual(result['cert_check']['init_config']['ciphers'],
                             'fake-cipher')
            self.assertEqual(result['cert_check']['init_config']['collect_period'],
                             1200)
            self.assertEqual(result['cert_check']['init_config']['timeout'],
                             1.0)
            self.assertTrue(mock_log.called)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 SUSE Linux GmbH
# Copyright 2017 OP5 AB
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Util functions to assist in detection.
"""
import argparse
import logging
import os
import pwd
import subprocess
from subprocess import CalledProcessError
from subprocess import PIPE
from subprocess import Popen

from oslo_config import cfg

from monasca_agent.common.psutil_wrapper import psutil
from monasca_setup import agent_config

log = logging.getLogger(__name__)

_DEFAULT_AGENT_USER = 'mon-agent'
_DETECTED_AGENT_USER = None

# check_output was introduced in python 2.7, function added
# to accommodate python 2.6
try:
    check_output = subprocess.check_output
except AttributeError:
    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be overridden.')
        process = Popen(stdout=PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise CalledProcessError(retcode, cmd)
        return output


def find_process_cmdline(search_string):
    """Simple function to search running process for one with cmdline
    containing search_string.
    """
    for process in psutil.process_iter():
        try:
            process_cmdline = ' '.join(process.as_dict(['cmdline'])['cmdline'])
            if (search_string in process_cmdline and
               'monasca-setup' not in process_cmdline):
                return process
        except psutil.NoSuchProcess:
            continue

    return None


def find_process_name(pname):
    """Simple function to search running process for one with pname.
    """
    for process in psutil.process_iter():
        try:
            if pname == process.as_dict(['name'])['name']:
                return process
        except psutil.NoSuchProcess:
            continue

    return None


def find_process_service(sname):
    """Simple function to call systemctl (service) to check if a service is running.
    """
    try:
        subprocess.check_call(['service', sname, 'status'], stdout=PIPE, stderr=PIPE)
        return True
    except subprocess.CalledProcessError:
        return False

    return False


def find_addrs_listening_on_port(port, kind='inet'):
    """Return the list of IP addresses which are listening
    on the specified port.
    """
    listening = []
    for connection in psutil.net_connections(kind):
        if (connection.status == psutil.CONN_LISTEN and
                connection.laddr[1] == int(port)):
            listening.append(connection.laddr[0])
    return listening


def find_addr_listening_on_port_over_tcp(port):
    """Return the IP address which is listening on the specified TCP port."""
    ip = find_addrs_listening_on_port(port, 'tcp')
    if ip:
        return ip[0].lstrip("::ffff:")


def get_agent_username():
    """Determine the user monasca-agent runs as"""

    global _DETECTED_AGENT_USER

    # Use cached agent user in subsequent calls
    if _DETECTED_AGENT_USER is not None:
        return _DETECTED_AGENT_USER

    # Try the owner of agent.yaml first
    try:
        uid = os.stat('/etc/monasca/agent/agent.yaml').st_uid
    except OSError:
        uid = None

    if uid is not None:
        _DETECTED_AGENT_USER = pwd.getpwuid(uid).pw_name
        return _DETECTED_AGENT_USER

    # No agent.yaml, so try to find a running monasca-agent process
    agent_process = find_process_name('monasca-agent')

    if agent_process is not None:
        _DETECTED_AGENT_USER = agent_process.username()
        return _DETECTED_AGENT_USER

    # Fall back to static agent user
    log.warn("Could not determine monasca-agent service user, falling "
             "back to %s" % _DEFAULT_AGENT_USER)
    _DETECTED_AGENT_USER = _DEFAULT_AGENT_USER
    return _DETECTED_AGENT_USER


def load_oslo_configuration(from_cmd, in_project,
                            for_opts, of_prog=None):
    """Loads configuration of an OpenStack project.

    for_opts should be a :py:class:`list` containing dictionaries
    with keys as expected by :py:class:meth:`cfg.ConfigOpts.register_opt`::

        >>> for_opts = [
        >>>     {'opt': cfg.StrOpt('region_name')},
        >>>     {'opt': cfg.StrOpt('username'), 'group': 'keystoneauth'},
        >>>     {'opt': cfg.StrOpt('password'), 'group': 'keystoneauth'},
        >>> ]

    Example::

        >>> nova_proc = find_process_name('nova-compute')
        >>> proc_cmd = nova_proc.as_dict(['cmdline'])['cmdline']
        >>> load_oslo_configuration(
        >>>     from_cmd=proc_cmd,
        >>>     in_project='nova',
        >>>     for_opts=for_opts
        >>> )

    which will load three [region_name, username and password] settings from
    Nova configuration regardless of where those
    settings are actually defined.

    :param from_cmd: cmdline of a process, used also to retrieve arguments
    :type from_cmd: list[basestring]
    :param in_project: the project name as defined in its oslo setup
    :type in_project: basestring
    :param for_opts: list of dict containing options to look for inside config
    :type for_opts: list[dict]
    :param of_prog: program name within the project [optional]
    :type of_prog: basestring
    :return: oslo configuration object
    :rtype: oslo_config.cfg.CONF
    """

    conf_holder = cfg.ConfigOpts()
    for no in for_opts:
        conf_holder.register_opt(**no)

    # NOTE(trebskit) we need to remove everything from the beginning
    # of the cmd arg list that is not an argument of the application
    # we want to get configuration from, i.e.;
    # /usr/bin/python, /usr/bin/python3
    # and next actual binary of the program
    # /usr/local/bin/nova-compute
    # NOTE(tobiajo) Just keep built-in options for oslo.config
    args = []
    parser = argparse.ArgumentParser()
    parser.add_argument('--config-file')
    parser.add_argument('--config-dir')
    namespace, _ = parser.parse_known_args(from_cmd[2:])
    if namespace.config_file:
        args.append('--config-file')
        args.append(namespace.config_file)
    if namespace.config_dir:
        args.append('--config-dir')
        args.append(namespace.config_dir)

    conf_holder(
        args=args,
        project=in_project,
        prog=of_prog
    )

    return conf_holder


def watch_process(search_strings, service=None, component=None,
                  exact_match=True, detailed=True, process_name=None, dimensions=None):
    """Takes a list of process search strings and returns a Plugins object with the config set.
        This was built as a helper as many plugins setup process watching
    """
    config = agent_config.Plugins()

    # Fallback to default process_name strategy if process_name is not defined
    process_name = process_name if process_name else search_strings[0]
    parameters = {'name': process_name,
                  'detailed': detailed,
                  'exact_match': exact_match,
                  'search_string': search_strings}

    dimensions = _get_dimensions(service, component, dimensions)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['process'] = {'init_config': None,
                         'instances': [parameters]}
    return config


def watch_process_by_username(username, process_name, service=None,
                              component=None, detailed=True, dimensions=None):
    """Takes a user and returns a Plugins object with the config set for a process check by user.
        This was built as a helper as many plugins setup process watching.
    """
    config = agent_config.Plugins()

    parameters = {'name': process_name,
                  'detailed': detailed,
                  'username': username}

    dimensions = _get_dimensions(service, component, dimensions)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['process'] = {'init_config': None,
                         'instances': [parameters]}
    return config


def watch_file_size(directory_name, file_names, file_recursive=False,
                    service=None, component=None, dimensions=None):
    """Takes a directory, a list of files, recursive flag and returns a
        Plugins object with the config set.
    """
    config = agent_config.Plugins()
    parameters = {'directory_name': directory_name,
                  'file_names': file_names,
                  'recursive': file_recursive}

    dimensions = _get_dimensions(service, component, dimensions)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['file_size'] = {'init_config': None,
                           'instances': [parameters]}
    return config


def watch_directory(directory_name, service=None, component=None, dimensions=None):
    """Takes a directory name and returns a Plugins object with the config set.
    """
    config = agent_config.Plugins()
    parameters = {'directory': directory_name}

    dimensions = _get_dimensions(service, component, dimensions)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['directory'] = {'init_config': None,
                           'instances': [parameters]}
    return config


def service_api_check(name, url, pattern, use_keystone=True,
                      service=None, component=None, dimensions=None):
    """Setup a service api to be watched by the http_check plugin.
    """
    config = agent_config.Plugins()
    parameters = {'name': name,
                  'url': url,
                  'match_pattern': pattern,
                  'timeout': 10,
                  'use_keystone': use_keystone}

    dimensions = _get_dimensions(service, component, dimensions)
    if len(dimensions) > 0:
        parameters['dimensions'] = dimensions

    config['http_check'] = {'init_config': None,
                            'instances': [parameters]}
    return config


def _get_dimensions(service, component, dimensions=None):
    if dimensions is None:
        dimensions = {}
    # If service parameter is set in the plugin config, add the service dimension which
    # will override the service in the agent config
    if service:
        dimensions.update({'service': service})

    # If component parameter is set in the plugin config, add the component dimension which
    # will override the component in the agent config
    if component:
        dimensions.update({'component': component})

    return dimensions



# (c) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
# Copyright 2017 SUSE Linux GmbH
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import yaml

from monasca_setup import agent_config
from monasca_setup.detection import plugin
from monasca_setup.detection import utils

log = logging.getLogger(__name__)

_POSTFIX_PROC_NAME = 'postfix'
_POSTFIX_DIRECTORY = """/var/spool/postfix"""
_POSTFIX_CHECK_COMMAND = ('sudo -l -U {0} find %s/incoming '
                          '-type f > /dev/null' % _POSTFIX_DIRECTORY)
"""Command to verify if user running monasca-agent
 has sudo permission to access postfix directory"""


class Postfix(plugin.Plugin):
    """If postfix is running install the default config.
    """

    ERROR_MSG = 'postfix plugin will not be configured.'

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """

        try:
            has_process = (utils.find_process_cmdline(_POSTFIX_PROC_NAME)
                           is not None)
            agent_user = utils.get_agent_username() if has_process else None
            has_user = agent_user is not None
            has_sudoers = (self._has_sudoers(agent_user)
                           if agent_user else False)
        except Exception:
            self.available = False
            detailed_msg = ('Unexpected exception while '
                            'running postfix detection.')
            log.exception('%s\n%s' % (detailed_msg, self.ERROR_MSG))
        else:
            self.available = has_process and has_sudoers
            if not self.available:
                if not has_process:
                    detailed_msg = ('%s process was not found.'
                                    % _POSTFIX_PROC_NAME)
                    log.info('%s\n%s' % (detailed_msg, self.ERROR_MSG))
                elif not has_user:
                    detailed_msg = 'Did not locate agent\'s username.'
                    log.error('%s\n%s' % (detailed_msg, self.ERROR_MSG))
                elif not has_sudoers:
                    detailed_msg = ('%s cannot access %s directory. '
                                    '\n Refer to postfix plugin documentation '
                                    'for more details.'
                                    % (agent_user, _POSTFIX_DIRECTORY))
                    log.warning('%s\n%s' % (detailed_msg, self.ERROR_MSG))

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        # A bit silly to parse the yaml only for it to be converted back but this
        # plugin is the exception not the rule
        with open(os.path.join(self.template_dir, 'conf.d/postfix.yaml.example'),
                  'r') as postfix_template:
            default_net_config = yaml.safe_load(postfix_template.read())
        config = agent_config.Plugins()
        config['postfix'] = default_net_config
        return config

    def dependencies_installed(self):
        return True

    @staticmethod
    def _has_sudoers(agent_user):
        test_sudo = os.system(_POSTFIX_CHECK_COMMAND.format(agent_user))
        return test_sudo == 0



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os
import subprocess
from unittest import mock
import unittest

from monasca_agent.common import util
from monasca_agent.collector.checks_d import ceph


def mocked_check_output(args, shell=True, stderr='', version='jewel'):
    output = ''
    if '-f json df detail' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-df.json'.format(version))
    elif '-f json status' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-status.json'.format(version))
    elif 'status' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-status.plain'.format(version))
    elif '-f json osd df' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-osd-df.json'.format(version))
    elif '-f json osd perf' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-osd-perf.json'.format(version))
    elif '-f json osd dump' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-osd-dump.json'.format(version))
    elif '-f json osd pool stats' in args:
        output = open(os.path.dirname(os.path.abspath(__file__)) +
                      '/fixtures/ceph-{}/test_ceph-osd-pool-stats.json'.format(version))
    else:
        raise subprocess.CalledProcessError(1, cmd=args,
                                            output='Invalid command')
    return output.read()

def mocked_check_output_jewel(args, shell=True, stderr=''):
    return mocked_check_output(args, shell, stderr, 'jewel')

def mocked_check_output_luminous(args, shell=True, stderr=''):
    return mocked_check_output(args, shell, stderr, 'luminous')

class MockCephCheck(ceph.Ceph):

    CLUSTER = 'ceph'

    def __init__(self, ceph_version='jewel'):

        # Attach representative output from Ceph clusters of different versions
        if ceph_version == 'jewel':
            subprocess.check_output = mock.Mock(
                side_effect=mocked_check_output_jewel)
        elif ceph_version == 'luminous':
            subprocess.check_output = mock.Mock(
                side_effect=mocked_check_output_luminous)

        super(MockCephCheck, self).__init__(
            name='ceph',
            init_config={},
            instances=[],
            agent_config={}
        )

    def _ceph_cmd(self, *args):
        if hasattr(self, 'instance'):
            return super(MockCephCheck, self)._ceph_cmd(*args)
        else:
            self.instance = { 'use_sudo': False }
            ret = super(MockCephCheck, self)._ceph_cmd(*args)
            del self.instance
            return ret


class CephCheckTest(unittest.TestCase):
    maxDiff = None

    CEPH_VERSION='jewel'

    expected = {
        'pool_stats': {
            'images': {
                'ceph.pool.recovery.recovering_objects_per_sec': 3530,
                'ceph.pool.recovery.recovering_bytes_per_sec': 14462655,
                'ceph.pool.recovery.recovering_keys_per_sec': 0,
                'ceph.pool.recovery.num_objects_recovered': 7148,
                'ceph.pool.recovery.num_bytes_recovered': 29278208,
                'ceph.pool.recovery.num_keys_recovered': 0
            },
            'vms': {
                'ceph.pool.client.read_bytes_sec': 16869,
                'ceph.pool.client.write_bytes_sec': 9341127,
                'ceph.pool.client.read_op_per_sec': 369,
                'ceph.pool.client.write_op_per_sec': 1364
            }
        },
        'df_total': {
            'ceph.cluster.total_bytes': 150000,
            'ceph.cluster.total_used_bytes': 90000,
            'ceph.cluster.total_avail_bytes': 60000,
            'ceph.cluster.objects.total_count': 50,
            'ceph.cluster.utilization_perc': 0.6
        },
        'stats_metrics': {
            'ceph.cluster.health_status': 0,
            'ceph.cluster.osds.total_count': 3,
            'ceph.cluster.osds.up_count': 3,
            'ceph.cluster.osds.in_count': 3,
            'ceph.cluster.osds.down_count': 0,
            'ceph.cluster.osds.out_count': 0,
            'ceph.cluster.pgs.degraded_count': 1,
            'ceph.cluster.pgs.stuck_unclean_count': 4,
            'ceph.cluster.pgs.undersized_count': 5,
            'ceph.cluster.objects.degraded_count': 10,
            'ceph.cluster.pgs.active+clean': 192,
            'ceph.cluster.pgs.active+clean+scrubbing+deep': 1,
            'ceph.cluster.pgs.active+clean+scrubbing': 1,
            'ceph.cluster.pgs.scrubbing_count': 1,
            'ceph.cluster.pgs.deep_scrubbing_count': 1,
            'ceph.cluster.pgs.remapped_count': 0,
            'ceph.cluster.pgs.total_count': 192,
            'ceph.cluster.pgs.avg_per_osd': 64,
            'ceph.cluster.client.read_bytes_per_sec': 630000.0,
            'ceph.cluster.client.write_bytes_per_sec': 272000000.0,
            'ceph.cluster.client.read_ops': 263,
            'ceph.cluster.client.write_ops': 1964,
            'ceph.cluster.recovery.bytes_per_sec': 1e9,
            'ceph.cluster.recovery.keys_per_sec': 100,
            'ceph.cluster.recovery.objects_per_sec': 50,
            'ceph.cluster.cache.flush_bytes_per_sec': 1e8,
            'ceph.cluster.cache.evict_bytes_per_sec': 1e9,
            'ceph.cluster.cache.promote_ops': 20,
            'ceph.cluster.quorum_size': 3
        },
        'osd_metrics': {
            'osd.0': {
                'ceph.osd.crush_weight': 0.999390,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.000000,
                'ceph.osd.total_bytes': 50000.0,
                'ceph.osd.used_bytes': 25000.0,
                'ceph.osd.avail_bytes': 25000.0,
                'ceph.osd.utilization_perc': 0.5,
                'ceph.osd.variance': 1.008811,
                'ceph.osd.pgs_count': 192,
                'ceph.osd.perf.commit_latency_seconds': 0.031,
                'ceph.osd.perf.apply_latency_seconds': 0.862,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.1': {
                'ceph.osd.crush_weight': 0.999390,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.000000,
                'ceph.osd.total_bytes': 50000.0,
                'ceph.osd.used_bytes': 25000.0,
                'ceph.osd.avail_bytes': 25000.0,
                'ceph.osd.utilization_perc': 0.5,
                'ceph.osd.variance': 0.998439,
                'ceph.osd.pgs_count': 192,
                'ceph.osd.perf.commit_latency_seconds': 0.025,
                'ceph.osd.perf.apply_latency_seconds': 1.390,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.2': {
                'ceph.osd.crush_weight': 0.999390,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.000000,
                'ceph.osd.total_bytes': 50000.0,
                'ceph.osd.used_bytes': 25000.0,
                'ceph.osd.avail_bytes': 25000.0,
                'ceph.osd.utilization_perc': 0.5,
                'ceph.osd.variance': 0.992750,
                'ceph.osd.pgs_count': 192,
                'ceph.osd.perf.commit_latency_seconds': 0.025,
                'ceph.osd.perf.apply_latency_seconds': 1.505,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            }
        },
        'summary_metrics': {
            'ceph.osds.total_bytes': 150000.0,
            'ceph.osds.total_used_bytes': 75000.0,
            'ceph.osds.total_avail_bytes': 75000.0,
            'ceph.osds.avg_utilization_perc': 0.5
        },
        'pool_metrics': {
            'images': {
                'ceph.pool.used_bytes': 10000,
                'ceph.pool.used_raw_bytes': 30000,
                'ceph.pool.max_avail_bytes': 20000,
                'ceph.pool.objects_count': 20,
                'ceph.pool.dirty_objects_count': 20,
                'ceph.pool.read_io': 6000,
                'ceph.pool.read_bytes': 20000,
                'ceph.pool.write_io': 2000,
                'ceph.pool.write_bytes': 20000,
                'ceph.pool.quota_max_bytes': 50000,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.total_bytes': 30000,
                'ceph.pool.utilization_perc': 0.3333333333333333
            },
            'vms': {
                'ceph.pool.used_bytes': 20000,
                'ceph.pool.used_raw_bytes': 60000,
                'ceph.pool.max_avail_bytes': 20000,
                'ceph.pool.objects_count': 30,
                'ceph.pool.dirty_objects_count': 30,
                'ceph.pool.read_io': 4000,
                'ceph.pool.read_bytes': 80000,
                'ceph.pool.write_io': 1000,
                'ceph.pool.write_bytes': 20000,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.total_bytes': 40000,
                'ceph.pool.utilization_perc': 0.5
            }
        },
        'mon_metrics': {
            'mon0': {
                'ceph.monitor.total_bytes': 100000.0,
                'ceph.monitor.used_bytes': 50000.0,
                'ceph.monitor.avail_bytes': 50000.0,
                'ceph.monitor.avail_perc': 50,
                'ceph.monitor.store.total_bytes': 100,
                'ceph.monitor.store.sst_bytes': 0,
                'ceph.monitor.store.log_bytes': 10,
                'ceph.monitor.store.misc_bytes': 10,
                'ceph.monitor.skew': 0.000000,
                'ceph.monitor.latency': 0.000000
            },
            'mon1': {
                'ceph.monitor.total_bytes': 100000.0,
                'ceph.monitor.used_bytes': 50000.0,
                'ceph.monitor.avail_bytes': 50000.0,
                'ceph.monitor.avail_perc': 50,
                'ceph.monitor.store.total_bytes': 100,
                'ceph.monitor.store.sst_bytes': 0,
                'ceph.monitor.store.log_bytes': 10,
                'ceph.monitor.store.misc_bytes': 10,
                'ceph.monitor.skew': 0.000000,
                'ceph.monitor.latency': 0.002577
            },
            'mon2': {
                'ceph.monitor.total_bytes': 100000.0,
                'ceph.monitor.used_bytes': 50000.0,
                'ceph.monitor.avail_bytes': 50000.0,
                'ceph.monitor.avail_perc': 50,
                'ceph.monitor.store.total_bytes': 100,
                'ceph.monitor.store.sst_bytes': 0,
                'ceph.monitor.store.log_bytes': 10,
                'ceph.monitor.store.misc_bytes': 10,
                'ceph.monitor.skew': 0.000000,
                'ceph.monitor.latency': 0.003353
            },
        },
        'test_ceph_cmd_lens': {
            'df detail json': 2,
            'status json': 9,
            'status': 683,
            'osd df json': 3,
            'osd perf json': 1,
            'osd dump json': 15,
            'osd pool stats json': 2
        },
        'test_check_call_count': 144
    }

    def setUp(self):
        super(CephCheckTest, self).setUp()
        self.ceph_check = MockCephCheck(self.CEPH_VERSION)
        self.ceph_check.gauge = mock.Mock()

    def test_ceph_cmd(self):
        df = self.ceph_check._ceph_cmd('df detail', 'json')
        st = self.ceph_check._ceph_cmd('status', 'json')
        st_plain = self.ceph_check._ceph_cmd('status')
        osd_df = self.ceph_check._ceph_cmd('osd df', 'json')
        osd_perf = self.ceph_check._ceph_cmd('osd perf', 'json')
        osd_dump = self.ceph_check._ceph_cmd('osd dump', 'json')
        osd_pool = self.ceph_check._ceph_cmd('osd pool stats', 'json')

        self.assertIsInstance(df, dict)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['df detail json'], len(df))
        self.assertIsInstance(st, dict)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['status json'], len(st))
        self.assertIsInstance(st_plain, str)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['status'], len(st_plain))
        self.assertIsInstance(osd_df, dict)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['osd df json'], len(osd_df))
        self.assertIsInstance(osd_perf, dict)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['osd perf json'], len(osd_perf))
        self.assertIsInstance(osd_dump, dict)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['osd dump json'], len(osd_dump))
        self.assertIsInstance(osd_pool, list)
        self.assertEqual(self.expected['test_ceph_cmd_lens']['osd pool stats json'], len(osd_pool))

        with self.assertRaises(subprocess.CalledProcessError) as e:
            self.ceph_check._ceph_cmd('foo', 'json')
            self.assertEqual("Unable to execute ceph command 'ceph --cluster"
                             "ceph -f json foo': Invalid command", e.output)

    def test_ceph_cmd_sudo(self):
        self.ceph_check.check({
            'use_sudo': True,
        })

        expect_cmd = 'sudo ceph --cluster ceph -f json df detail'

        with mock.patch('subprocess.check_output') as ceph_cmd_call:
            try:
                self.ceph_check._ceph_cmd('df detail', 'json')
            except Exception as e:
                pass
            ceph_cmd_call.assert_called_with(expect_cmd, shell=True,
                                             stderr=subprocess.STDOUT)

    def test_parse_ceph_status(self):
        self.assertEqual(0, self.ceph_check._parse_ceph_status('HEALTH_OK'))
        self.assertEqual(1, self.ceph_check._parse_ceph_status('HEALTH_WARN'))
        self.assertEqual(2, self.ceph_check._parse_ceph_status('HEALTH_ERR'))
        self.assertEqual(2, self.ceph_check._parse_ceph_status('foo'))

    def test_get_cache_io(self):
        cache_kb = 'cache io 1000000 kB/s flush, 1000000 kB/s evict,' \
                   ' 20 op/s promote'
        cache_mb = 'cache io 1000 MB/s flush, 1000 MB/s evict, 20 op/s promote'
        cache_gb = 'cache io 1 GB/s flush, 1 GB/s evict, 20 op/s promote'
        expected_metrics = {
            'ceph.cluster.cache.flush_bytes_per_sec': 1e9,
            'ceph.cluster.cache.evict_bytes_per_sec': 1e9,
            'ceph.cluster.cache.promote_ops': 20
        }

        metrics_kb = self.ceph_check._get_cache_io(cache_kb)
        metrics_mb = self.ceph_check._get_cache_io(cache_mb)
        metrics_gb = self.ceph_check._get_cache_io(cache_gb)
        self.assertEqual(expected_metrics, metrics_kb)
        self.assertEqual(expected_metrics, metrics_mb)
        self.assertEqual(expected_metrics, metrics_gb)

    def test_get_client_io(self):
        client_kb = 'client io 1000000 kB/s rd, 1000000 kb/s wr, 10 op/s rd,' \
                     ' 20 op/s wr'
        client_mb = 'client io 1000 MB/s rd, 1000 mb/s wr, 10 op/s rd,' \
                    ' 20 op/s wr'
        client_gb = 'client io 1 GB/s rd, 1 gb/s wr, 10 op/s rd, 20 op/s wr'
        expected_metrics = {
            'ceph.cluster.client.read_bytes_per_sec': 1e9,
            'ceph.cluster.client.write_bytes_per_sec': 1e9,
            'ceph.cluster.client.read_ops': 10,
            'ceph.cluster.client.write_ops': 20
        }

        metrics_kb = self.ceph_check._get_client_io(client_kb)
        metrics_mb = self.ceph_check._get_client_io(client_mb)
        metrics_gb = self.ceph_check._get_client_io(client_gb)
        self.assertEqual(expected_metrics, metrics_kb)
        self.assertEqual(expected_metrics, metrics_mb)
        self.assertEqual(expected_metrics, metrics_gb)

    def test_get_recovery_io(self):
        recovery_kb = 'recovery io 1000000 kB/s, 100 keys/s, 50 objects/s'
        recovery_mb = 'recovery io 1000 MB/s, 100 keys/s, 50 objects/s'
        recovery_gb = 'recovery io 1 GB/s, 100 keys/s, 50 objects/s'
        expected_metrics = {
            'ceph.cluster.recovery.bytes_per_sec': 1e9,
            'ceph.cluster.recovery.keys_per_sec': 100,
            'ceph.cluster.recovery.objects_per_sec': 50
        }

        metrics_kb = self.ceph_check._get_recovery_io(recovery_kb)
        metrics_mb = self.ceph_check._get_recovery_io(recovery_mb)
        metrics_gb = self.ceph_check._get_recovery_io(recovery_gb)
        self.assertEqual(expected_metrics, metrics_kb)
        self.assertEqual(expected_metrics, metrics_mb)
        self.assertEqual(expected_metrics, metrics_gb)

    def test_get_summary_metrics(self):
        summary_strs = [
            '1 pgs degraded', '2 pgs stuck degraded', '3 pgs unclean',
            '4 pgs stuck unclean', '5 pgs undersized',
            '6 pgs stuck undersized', '7 pgs stale', '8 pgs stuck stale',
            '9 requests are blocked', 'recovery 10/100 objects degraded',
            'recovery 11/100 objects misplaced'
        ]

        expected_metrics = {
            'ceph.cluster.pgs.degraded_count': 1,
            'ceph.cluster.pgs.stuck_degraded_count': 2,
            'ceph.cluster.pgs.unclean_count': 3,
            'ceph.cluster.pgs.stuck_unclean_count': 4,
            'ceph.cluster.pgs.undersized_count': 5,
            'ceph.cluster.pgs.stuck_undersized_count': 6,
            'ceph.cluster.pgs.stale_count': 7,
            'ceph.cluster.pgs.stuck_stale_count': 8,
            'ceph.cluster.slow_requests_count': 9,
            'ceph.cluster.objects.degraded_count': 10,
            'ceph.cluster.objects.misplaced_count': 11
        }

        metrics = {}
        self.assertEqual(self.ceph_check._get_summary_metrics(''), {})
        for s in summary_strs:
            metrics.update(self.ceph_check._get_summary_metrics(s))
        self.assertEqual(expected_metrics, metrics)

    def test_get_usage_metrics(self):
        df = self.ceph_check._ceph_cmd('df detail', 'json')
        metrics = self.ceph_check._get_usage_metrics(df)

        # NOTE: some metrics are floating point values and may not be
        # reliably compared with pre-computed float values
        self.assertEqual(
            { k: v for k, v in self.expected['df_total'].items()
                if k != 'ceph.cluster.utilization_perc'},
            { k: v for k, v in metrics.items()
                if k != 'ceph.cluster.utilization_perc'})
        self.assertAlmostEqual(
            self.expected['df_total']['ceph.cluster.utilization_perc'],
            metrics['ceph.cluster.utilization_perc'], places=3)

    def test_get_stats_metrics(self):
        status = self.ceph_check._ceph_cmd('status', 'json')
        status_plain = self.ceph_check._ceph_cmd('status')
        metrics = self.ceph_check._get_stats_metrics(status, status_plain)
        self.assertDictEqual(self.expected['stats_metrics'], metrics)

    def test_get_mon_metrics(self):
        status = self.ceph_check._ceph_cmd('status', 'json')

        metrics = self.ceph_check._get_mon_metrics(status)
        self.assertEqual(self.expected['mon_metrics'], metrics)

    def test_get_osd_metrics(self):
        df = self.ceph_check._ceph_cmd('osd df', 'json')
        perf = self.ceph_check._ceph_cmd('osd perf', 'json')
        dump = self.ceph_check._ceph_cmd('osd dump', 'json')

        metrics = self.ceph_check._get_osd_metrics(df, perf, dump)

        # NOTE: some metrics are floating point values and may not be
        # reliably compared with pre-computed float values
        for osd, osd_metrics in metrics.items():
            self.assertEqual(
                { k: v for k, v in self.expected['osd_metrics'][osd].items()
                    if not k in ['ceph.osd.utilization_perc', 'ceph.osd.variance']},
                { k: v for k, v in osd_metrics.items()
                    if not k in ['ceph.osd.utilization_perc', 'ceph.osd.variance']})

            self.assertAlmostEqual(
                self.expected['osd_metrics'][osd]['ceph.osd.utilization_perc'],
                osd_metrics['ceph.osd.utilization_perc'], places=2)
            self.assertAlmostEqual(
                self.expected['osd_metrics'][osd]['ceph.osd.variance'],
                osd_metrics['ceph.osd.variance'], places=2)

    def test_get_osd_summary_metrics(self):
        df = self.ceph_check._ceph_cmd('osd df', 'json')

        metrics = self.ceph_check._get_osd_summary_metrics(df)

        # NOTE: some metrics are floating point values and may not be
        # reliably compared with pre-computed float values
        self.assertEqual(
            { k: v for k, v in self.expected['summary_metrics'].items()
                if k != 'ceph.osds.avg_utilization_perc'},
            { k: v for k, v in metrics.items()
                if k != 'ceph.osds.avg_utilization_perc'})
        self.assertAlmostEqual(
            self.expected['summary_metrics']['ceph.osds.avg_utilization_perc'],
            metrics['ceph.osds.avg_utilization_perc'], places=3)


    def test_get_pool_metrics(self):
        df = self.ceph_check._ceph_cmd('df detail', 'json')

        # NOTE: some metrics are floating point values and may not be
        # reliably compared with pre-computed float values
        metrics = self.ceph_check._get_pool_metrics(df)
        for pool, pool_metrics in metrics.items():
            self.assertEqual(
                { k: v for k, v in self.expected['pool_metrics'][pool].items()
                    if k != 'ceph.pool.utilization_perc'},
                { k: v for k, v in metrics[pool].items()
                    if k != 'ceph.pool.utilization_perc'})

        self.assertAlmostEqual(
            self.expected['pool_metrics'][pool]['ceph.pool.utilization_perc'],
            pool_metrics['ceph.pool.utilization_perc'], places=3)

    def test_get_pool_stats_metrics(self):
        pool_stats = self.ceph_check._ceph_cmd('osd pool stats', 'json')
        metrics = self.ceph_check._get_pool_stats_metrics(pool_stats)

        self.assertEqual(self.expected['pool_stats'], metrics)

    def test_check(self):
        self.ceph_check.check({})
        self.assertEqual(self.expected['test_check_call_count'], self.ceph_check.gauge.call_count)

    def test_check_disable_all_metrics(self):
        self.ceph_check._get_usage_metrics = mock.Mock(return_value={})
        self.ceph_check._get_stats_metrics = mock.Mock(return_value={})
        self.ceph_check._get_mon_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_summary_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_stats_metrics = mock.Mock(return_value={})

        self.ceph_check.check({
            'collect_usage_metrics': False,
            'collect_stats_metrics': False,
            'collect_mon_metrics': False,
            'collect_osd_metrics': False,
            'collect_pool_metrics': False,
        })

        self.assertFalse(self.ceph_check._get_usage_metrics.called)
        self.assertFalse(self.ceph_check._get_stats_metrics.called)
        self.assertFalse(self.ceph_check._get_mon_metrics.called)
        self.assertFalse(self.ceph_check._get_osd_metrics.called)
        self.assertFalse(self.ceph_check._get_osd_summary_metrics.called)
        self.assertFalse(self.ceph_check._get_pool_metrics.called)
        self.assertFalse(self.ceph_check._get_pool_stats_metrics.called)

    def test_check_disable_some_metrics(self):
        self.ceph_check._get_usage_metrics = mock.Mock(return_value={})
        self.ceph_check._get_stats_metrics = mock.Mock(return_value={})
        self.ceph_check._get_mon_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_summary_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_stats_metrics = mock.Mock(return_value={})

        self.ceph_check.check({
            'collect_usage_metrics': False,
            'collect_stats_metrics': False
        })

        self.assertFalse(self.ceph_check._get_usage_metrics.called)
        self.assertFalse(self.ceph_check._get_stats_metrics.called)
        self.assertTrue(self.ceph_check._get_mon_metrics.called)
        self.assertTrue(self.ceph_check._get_osd_metrics.called)
        self.assertTrue(self.ceph_check._get_osd_summary_metrics.called)
        self.assertTrue(self.ceph_check._get_pool_metrics.called)
        self.assertTrue(self.ceph_check._get_pool_stats_metrics.called)

    def test_check_enable_all_metrics(self):
        self.ceph_check._get_usage_metrics = mock.Mock(return_value={})
        self.ceph_check._get_stats_metrics = mock.Mock(return_value={})
        self.ceph_check._get_mon_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_metrics = mock.Mock(return_value={})
        self.ceph_check._get_osd_summary_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_metrics = mock.Mock(return_value={})
        self.ceph_check._get_pool_stats_metrics = mock.Mock(return_value={})

        self.ceph_check.check({
            'collect_usage_metrics': True,
            'collect_stats_metrics': True,
            'collect_mon_metrics': True,
            'collect_osd_metrics': True,
            'collect_pool_metrics': True,
        })

        self.assertTrue(self.ceph_check._get_usage_metrics.called)
        self.assertTrue(self.ceph_check._get_stats_metrics.called)
        self.assertTrue(self.ceph_check._get_mon_metrics.called)
        self.assertTrue(self.ceph_check._get_osd_metrics.called)
        self.assertTrue(self.ceph_check._get_osd_summary_metrics.called)
        self.assertTrue(self.ceph_check._get_pool_metrics.called)
        self.assertTrue(self.ceph_check._get_pool_stats_metrics.called)


class CephCheckTestLuminous(CephCheckTest):

    CEPH_VERSION='luminous'

    expected = {
        'pool_stats': {
        },
        'df_total': {
            'ceph.cluster.total_bytes': 121070163910656,
            'ceph.cluster.total_used_bytes': 7961411256320,
            'ceph.cluster.total_avail_bytes': 113108752654336,
            'ceph.cluster.objects.total_count': 34566782,
            'ceph.cluster.utilization_perc': 0.0658
        },
        'stats_metrics': {
            'ceph.cluster.health_status': 1,
            'ceph.cluster.osds.total_count': 33,
            'ceph.cluster.osds.up_count': 22,
            'ceph.cluster.osds.in_count': 24,
            'ceph.cluster.osds.down_count': 11,
            'ceph.cluster.osds.out_count': 9,
            #'ceph.cluster.pgs.degraded_count': 2361,
            #'ceph.cluster.pgs.stuck_unclean_count': 2361,
            #'ceph.cluster.pgs.undersized_count': 2361,
            #'ceph.cluster.objects.degraded_count': 9196074,
            'ceph.cluster.pgs.active+clean': 873,
            'ceph.cluster.pgs.active+undersized+degraded': 2361,
            'ceph.cluster.pgs.active+clean+remapped': 14,
            'ceph.cluster.pgs.scrubbing_count': 0,
            'ceph.cluster.pgs.deep_scrubbing_count': 0,
            'ceph.cluster.pgs.remapped_count': 14,
            'ceph.cluster.pgs.total_count': 3248,
            'ceph.cluster.pgs.avg_per_osd': 98.42424242424242,
            #'ceph.cluster.client.read_bytes_per_sec': 4246.0,
            'ceph.cluster.client.write_bytes_per_sec': 374000.0,
            'ceph.cluster.client.read_ops': 2,
            'ceph.cluster.client.write_ops': 16,
            #'ceph.cluster.recovery.bytes_per_sec': 0,
            #'ceph.cluster.recovery.keys_per_sec': 0,
            #'ceph.cluster.recovery.objects_per_sec': 0,
            'ceph.cluster.quorum_size': 3,
        },
        'osd_metrics': {
            'osd.0': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 407044864000.0,
                'ceph.osd.avail_bytes': 5479588668000.0,
                'ceph.osd.utilization_perc': 6.91,
                'ceph.osd.variance': 1.051531,
                'ceph.osd.pgs_count': 228,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.1': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 411742144000.0,
                'ceph.osd.avail_bytes': 5474891388000.0,
                'ceph.osd.utilization_perc': 6.99,
                'ceph.osd.variance': 1.063665,
                'ceph.osd.pgs_count': 204,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.2': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 442908160000.0,
                'ceph.osd.avail_bytes': 5443725372000.0,
                'ceph.osd.utilization_perc': 7.52,
                'ceph.osd.variance': 1.144177,
                'ceph.osd.pgs_count': 227,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.3': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 348767744000.0,
                'ceph.osd.avail_bytes': 5537865788000.0,
                'ceph.osd.utilization_perc': 5.92,
                'ceph.osd.variance': 0.900982,
                'ceph.osd.pgs_count': 196,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.4': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 383320704000.0,
                'ceph.osd.avail_bytes': 5503312828000.0,
                'ceph.osd.utilization_perc': 6.51,
                'ceph.osd.variance': 0.990243,
                'ceph.osd.pgs_count': 204,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.5': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 356644416000.0,
                'ceph.osd.avail_bytes': 5529989116000.0,
                'ceph.osd.utilization_perc': 6.06,
                'ceph.osd.variance': 0.92133,
                'ceph.osd.pgs_count': 197,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.6': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 325735936000.0,
                'ceph.osd.avail_bytes': 5560897596000.0,
                'ceph.osd.utilization_perc': 5.53,
                'ceph.osd.variance': 0.841483,
                'ceph.osd.pgs_count': 185,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.7': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 367045952000.0,
                'ceph.osd.avail_bytes': 5519587580000.0,
                'ceph.osd.utilization_perc': 6.24,
                'ceph.osd.variance': 0.9482,
                'ceph.osd.pgs_count': 205,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.8': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 394521920000.0,
                'ceph.osd.avail_bytes': 5492111612000.0,
                'ceph.osd.utilization_perc': 6.7,
                'ceph.osd.variance': 1.01918,
                'ceph.osd.pgs_count': 214,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.9': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 369788800000.0,
                'ceph.osd.avail_bytes': 5516844732000.0,
                'ceph.osd.utilization_perc': 6.28,
                'ceph.osd.variance': 0.955286,
                'ceph.osd.pgs_count': 206,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.10': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 427759872000.0,
                'ceph.osd.avail_bytes': 5458873660000.0,
                'ceph.osd.utilization_perc': 7.27,
                'ceph.osd.variance': 1.105044,
                'ceph.osd.pgs_count': 218,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.11': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 404752640000.0,
                'ceph.osd.avail_bytes': 5481880892000.0,
                'ceph.osd.utilization_perc': 6.88,
                'ceph.osd.variance': 1.045609,
                'ceph.osd.pgs_count': 194,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.12': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 437294592000.0,
                'ceph.osd.avail_bytes': 5449338940000.0,
                'ceph.osd.utilization_perc': 7.43,
                'ceph.osd.variance': 1.129676,
                'ceph.osd.pgs_count': 230,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.13': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 388802944000.0,
                'ceph.osd.avail_bytes': 5497830588000.0,
                'ceph.osd.utilization_perc': 6.6,
                'ceph.osd.variance': 1.004406,
                'ceph.osd.pgs_count': 218,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.14': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 330687424000.0,
                'ceph.osd.avail_bytes': 5555946108000.0,
                'ceph.osd.utilization_perc': 5.62,
                'ceph.osd.variance': 0.854274,
                'ceph.osd.pgs_count': 190,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.15': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 405913792000.0,
                'ceph.osd.avail_bytes': 5480719740000.0,
                'ceph.osd.utilization_perc': 6.9,
                'ceph.osd.variance': 1.048609,
                'ceph.osd.pgs_count': 225,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.16': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 377519552000.0,
                'ceph.osd.avail_bytes': 5509113980000.0,
                'ceph.osd.utilization_perc': 6.41,
                'ceph.osd.variance': 0.975257,
                'ceph.osd.pgs_count': 192,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.17': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 362563072000.0,
                'ceph.osd.avail_bytes': 5524070460000.0,
                'ceph.osd.utilization_perc': 6.16,
                'ceph.osd.variance': 0.936619,
                'ceph.osd.pgs_count': 199,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.18': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 403252800000.0,
                'ceph.osd.avail_bytes': 5483380732000.0,
                'ceph.osd.utilization_perc': 6.85,
                'ceph.osd.variance': 1.041734,
                'ceph.osd.pgs_count': 229,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.19': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 5886633532000.0,
                'ceph.osd.used_bytes': 416788096000.0,
                'ceph.osd.avail_bytes': 5469845436000.0,
                'ceph.osd.utilization_perc': 7.08,
                'ceph.osd.variance': 1.076701,
                'ceph.osd.pgs_count': 210,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.20': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.21': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.22': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.23': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.24': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.25': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.26': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.27': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
            'osd.28': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 1
            },
            'osd.29': {
                'ceph.osd.crush_weight': 5.482391,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 1
            },
            'osd.30': {
                'ceph.osd.crush_weight': 0.232788,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 249955652000.0,
                'ceph.osd.used_bytes': 6512248000.0,
                'ceph.osd.avail_bytes': 243443404000.0,
                'ceph.osd.utilization_perc': 2.61,
                'ceph.osd.variance': 0.3962,
                'ceph.osd.pgs_count': 1030,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.31': {
                'ceph.osd.crush_weight': 0.232788,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 1.0,
                'ceph.osd.total_bytes': 249955652000.0,
                'ceph.osd.used_bytes': 5456072000.0,
                'ceph.osd.avail_bytes': 244499580000.0,
                'ceph.osd.utilization_perc': 2.18,
                'ceph.osd.variance': 0.331943,
                'ceph.osd.pgs_count': 1030,
                'ceph.osd.perf.commit_latency_seconds': 0.0,
                'ceph.osd.perf.apply_latency_seconds': 0.0,
                'ceph.osd.up': 1,
                'ceph.osd.in': 1
            },
            'osd.32': {
                'ceph.osd.crush_weight': 0.232788,
                'ceph.osd.depth': 2,
                'ceph.osd.reweight': 0.0,
                'ceph.osd.total_bytes': 0.0,
                'ceph.osd.used_bytes': 0.0,
                'ceph.osd.avail_bytes': 0.0,
                'ceph.osd.utilization_perc': 0.0,
                'ceph.osd.variance': 0.0,
                'ceph.osd.pgs_count': 0,
                'ceph.osd.up': 0,
                'ceph.osd.in': 0
            },
        },
        'summary_metrics': {
            'ceph.osds.total_bytes': 118232581944000.0,
            'ceph.osds.total_used_bytes': 7774823744000.0,
            'ceph.osds.total_avail_bytes': 110457758200000.0,
            'ceph.osds.avg_utilization_perc': 6.575872
        },
        'pool_metrics': {
            'cephfs_metadata': {
                'ceph.pool.dirty_objects_count': 66133,
                'ceph.pool.max_avail_bytes': 236492341248,
                'ceph.pool.objects_count': 66133,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.read_bytes': 1511467008,
                'ceph.pool.read_io': 545015,
                'ceph.pool.used_bytes': 47653204,
                'ceph.pool.used_raw_bytes': 91122080,
                'ceph.pool.utilization_perc': 0.000201459394257617,
                'ceph.pool.total_bytes': 236539994452,
                'ceph.pool.write_io': 1823016,
                'ceph.pool.write_bytes': 15099714560
            },
            'cephfs_data': {
                'ceph.pool.used_bytes': 1743163949040,
                'ceph.pool.dirty_objects_count': 1287159,
                'ceph.pool.max_avail_bytes': 52729787449344,
                'ceph.pool.objects_count': 1287159,
                'ceph.pool.utilization_perc': 0.0320005416319651,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 3419234369536,
                'ceph.pool.read_io': 624295721,
                'ceph.pool.read_bytes': 25137454158848,
                'ceph.pool.total_bytes': 54472951398384,
                'ceph.pool.write_io': 65020292,
                'ceph.pool.write_bytes': 18614600047616
            },
            '.rgw.root': {
                'ceph.pool.used_bytes': 1113,
                'ceph.pool.dirty_objects_count': 4,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 4,
                'ceph.pool.utilization_perc': 0.0000000000210183661709385,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 2226,
                'ceph.pool.read_io': 4806,
                'ceph.pool.read_bytes': 3280896,
                'ceph.pool.total_bytes': 52953687786585,
                'ceph.pool.write_io': 4,
                'ceph.pool.write_bytes': 4096
            },
            'default.rgw.control': {
                'ceph.pool.used_bytes': 0,
                'ceph.pool.dirty_objects_count': 8,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 8,
                'ceph.pool.utilization_perc': 0.0,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 0,
                'ceph.pool.read_io': 0,
                'ceph.pool.read_bytes': 0,
                'ceph.pool.total_bytes': 52953687785472,
                'ceph.pool.write_io': 0,
                'ceph.pool.write_bytes': 0
            },
            'default.rgw.meta': {
                'ceph.pool.used_bytes': 832,
                'ceph.pool.dirty_objects_count': 6,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 6,
                'ceph.pool.utilization_perc': 0.0000000000157118424567059,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 1664,
                'ceph.pool.read_io': 220,
                'ceph.pool.read_bytes': 204800,
                'ceph.pool.total_bytes': 52953687786304,
                'ceph.pool.write_io': 147,
                'ceph.pool.write_bytes': 19456
            },
            'default.rgw.log': {
                'ceph.pool.used_bytes': 0,
                'ceph.pool.dirty_objects_count': 207,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 207,
                'ceph.pool.utilization_perc': 0.0,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 0,
                'ceph.pool.read_io': 21378199,
                'ceph.pool.read_bytes': 21891063808,
                'ceph.pool.total_bytes': 52953687785472,
                'ceph.pool.write_io': 14245040,
                'ceph.pool.write_bytes': 0
            },
            'default.rgw.buckets.index': {
                'ceph.pool.used_bytes': 0,
                'ceph.pool.dirty_objects_count': 1,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 1,
                'ceph.pool.utilization_perc': 0.0,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 0,
                'ceph.pool.read_io': 562,
                'ceph.pool.read_bytes': 580608,
                'ceph.pool.total_bytes': 52953687785472,
                'ceph.pool.write_io': 237,
                'ceph.pool.write_bytes': 0
            },
            'default.rgw.buckets.data': {
                'ceph.pool.used_bytes': 95250,
                'ceph.pool.dirty_objects_count': 6,
                'ceph.pool.max_avail_bytes': 52953687785472,
                'ceph.pool.objects_count': 6,
                'ceph.pool.utilization_perc': 0.00000000179891153595516,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 190500,
                'ceph.pool.read_io': 162,
                'ceph.pool.read_bytes': 291840,
                'ceph.pool.total_bytes': 52953687880722,
                'ceph.pool.write_io': 427,
                'ceph.pool.write_bytes': 602112
            },
            'kubernetes': {
                'ceph.pool.used_bytes': 4859,
                'ceph.pool.dirty_objects_count': 8,
                'ceph.pool.max_avail_bytes': 79430529581056,
                'ceph.pool.objects_count': 8,
                'ceph.pool.utilization_perc': 0.0000000000611729523311855,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 8503,
                'ceph.pool.read_io': 164,
                'ceph.pool.read_bytes': 133120,
                'ceph.pool.total_bytes': 79430529585915,
                'ceph.pool.write_io': 90892,
                'ceph.pool.write_bytes': 126869424128
            },
            'volumes': {
                'ceph.pool.used_bytes': 0,
                'ceph.pool.dirty_objects_count': 0,
                'ceph.pool.max_avail_bytes': 79430529581056,
                'ceph.pool.objects_count': 0,
                'ceph.pool.utilization_perc': 0.0,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 0,
                'ceph.pool.read_io': 0,
                'ceph.pool.read_bytes': 0,
                'ceph.pool.total_bytes': 79430529581056,
                'ceph.pool.write_io': 0,
                'ceph.pool.write_bytes': 0
            },
            'benchmark': {
                'ceph.pool.used_bytes': 37906503270,
                'ceph.pool.dirty_objects_count': 33213250,
                'ceph.pool.max_avail_bytes': 79430529581056,
                'ceph.pool.objects_count': 33213250,
                'ceph.pool.utilization_perc': 0.000477000745676893,
                'ceph.pool.quota_max_bytes': 0,
                'ceph.pool.quota_max_objects': 0,
                'ceph.pool.used_raw_bytes': 66925428736,
                'ceph.pool.read_io': 2261777,
                'ceph.pool.read_bytes': 3829217280,
                'ceph.pool.total_bytes': 79468436084326,
                'ceph.pool.write_io': 33280403,
                'ceph.pool.write_bytes': 62774775808
            }
        },
        'mon_metrics': {
        },
        'test_ceph_cmd_lens': {
            'df detail json': 2,
            'status json': 11,
            'status': 842,
            'osd df json': 3,
            'osd perf json': 1,
            'osd dump json': 24,
            'osd pool stats json': 11
        },
        'test_check_call_count': 578
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
from io import StringIO
import unittest

from collector.checks import LaconicFilter



class TestLaconic(unittest.TestCase):

    """Verify that we only output messages once
    """

    def setUp(self):
        self.l = logging.getLogger("test_laconic")
        self.sio = StringIO()

        # create console handler and set level to debug
        ch = logging.StreamHandler(self.sio)
        ch.setLevel(logging.DEBUG)

        # create formatter
        formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.l.addHandler(ch)

        self.laconic = LaconicFilter()
        self.l.addFilter(self.laconic)

    def tearDown(self):
        del self.l
        del self.sio

    def testRepeatingErrors(self):
        for i in range(10):
            self.l.error("Cannot find nagios.log")
        self.assertEqual(
            self.sio.getvalue().count("Cannot find nagios.log"), 1, self.sio.getvalue())

        for i in range(10):
            self.l.warn("Cannot find ganglia.log")
        self.assertEqual(
            self.sio.getvalue().count("Cannot find ganglia.log"), 1, self.sio.getvalue())

        for i in range(10):
            try:
                raise Exception("Ka-boom")
            except Exception:
                self.l.exception("Caught!")

        # once for the traceback, once for the message
        self.assertEqual(self.sio.getvalue().count("Ka-boom"), 2)

    def testNonRepeat(self):
        for i in range(10):
            self.l.error("Cannot find nagios.log %d" % i)
        self.assertEqual(self.sio.getvalue().count(" nagios.log"), 10)
        self.assertEqual(self.sio.getvalue().count(" 7"), 1)

    def testBlowUp(self):
        """Try to use a lot of memory"""
        for i in range(2 * self.laconic.LACONIC_MEM_LIMIT + 7):
            self.l.warn("""%d Lorem ipsum dolor sit amet, consectetur adipiscing elit. Integer
 adipiscing urna sed urna sagittis rutrum. Etiam elementum justo quis
 quam pellentesque vel porta eros vestibulum. Nulla nec felis
 lacus. Maecenas sit amet aliquet nunc. Curabitur non tellus eget
 sapien faucibus feugiat. Vivamus tortor nisl, posuere eget
 ullamcorper at, interdum id odio. Sed convallis, nisl quis luctus
 posuere, nulla sapien consequat magna, vitae euismod mauris felis et
 augue. Maecenas eget tortor elit. Cras eu nulla sit amet est
 hendrerit malesuada sit amet eu nibh. Aliquam erat volutpat. Cum
 sociis natoque penatibus et magnis dis parturient montes, nascetur
 ridiculus mus. Proin et erat diam, non venenatis dui. Etiam feugiat
 mattis nunc, tincidunt mollis quam condimentum eu. Donec volutpat
 sodales magna eu fermentum. Integer ultricies odio non metus aliquet
 tristique. Proin ultrices accumsan augue, quis tempor diam rutrum at.""" % i)
        self.assertEqual(len(self.laconic.hashed_messages), 7, self.sio.getvalue())

if __name__ == "__main__":
    unittest.main()
